import logging
import os

from nameko import config
from nameko.containers import WorkerContext
from nameko.extensions import DependencyProvider
from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from pyrabbit.api import Client
from raven import Client as RavenClient
from raven.conf import setup_logging
from raven.handlers.logging import SentryHandler

from .discovery import ServiceDiscovery


class ContainerProvider(DependencyProvider):
    """
    Allows access to ``ServiceContainer`` running current worker.
    """

    def get_dependency(self, worker_ctx):
        """
        Returns a ``ServiceContainer`` instance which runs current worker.
        """
        return self.container


class SentryLoggerConfig(DependencyProvider):
    def setup(self):
        sentry_config = config.get("SENTRY", {})
        dsn = sentry_config.get("DSN", None)
        if dsn:
            client = RavenClient(
                dsn, environment=os.environ.get("SENTRY_ENVIRONMENT", "local")
            )
            handler = SentryHandler(client)
            handler.setLevel(logging.ERROR)
            setup_logging(handler)


class ServiceDiscoveryProvider(DependencyProvider):
    def __init__(self, management_host: str, username: str, password: str):
        self.management_host = management_host
        self.username = username
        self.password = password

    def get_dependency(self, worker_ctx: WorkerContext) -> ServiceDiscovery:
        client = Client(self.management_host, self.username, self.password)
        return ServiceDiscovery(client)


class OpenTelemetryConfig(DependencyProvider):
    """
    Configures OTel trace exporter over HTTP.
    """

    def setup(self):
        resource = Resource(attributes={SERVICE_NAME: self.container.service_name})
        provider = TracerProvider(resource=resource)
        processor = BatchSpanProcessor(OTLPSpanExporter())
        provider.add_span_processor(processor)
        trace.set_tracer_provider(provider)



import pytest

try:
    import rich  # noqa

    has_rich = True
except ImportError:
    has_rich = False

from nameko_chassis.debug import debug_state_rich, debug_state_simple
from nameko_chassis.service import ServiceState, WorkerState


@pytest.fixture
def service_state() -> ServiceState:
    worker_state = WorkerState(
        class_name="MyService",
        method_name="my_method",
        args=["spam"],
        kwargs={"answer": "42"},
        data={},
        stacktrace=[],
    )
    return ServiceState(
        version="unknown",
        service_name="my_service",
        uptime=60,
        entrypoints=["my_method"],
        dependencies=["container"],
        running_workers=1,
        max_workers=10,
        worker_states=[worker_state],
    )


def test_debug_state_simple_prints_worker_availability(service_state, capsys) -> None:
    debug_state_simple(service_state)
    captured = capsys.readouterr()
    assert service_state.service_name in captured.out
    assert "Running 1/10 worker threads" in captured.out


@pytest.mark.skipif(not has_rich, reason="rich is not installed")
def test_debug_state_rich_prints_worker_availability(service_state, capsys) -> None:
    debug_state_rich(service_state)
    captured = capsys.readouterr()
    assert service_state.service_name in captured.out
    assert "Running 1/10 worker threads" in captured.out



import logging

import eventlet
from nameko.rpc import Client

logger = logging.getLogger(__name__)


class ServiceTimeout(Exception):
    pass


def is_service_responsive(
    service_proxy: Client,
    fail_gracefully=False,
    timeout: float = 5,
    method_name: str = "say_hello",
) -> bool:
    """
    A poor man's circuit breaker for nameko service proxies.

    True circuit breaker would wrap each and every RPC method and monitor
    it's error rate and duration. This implementation only checks if the
    service responds within ``timeout`` seconds. By default, it raises
    an exception if service is unreachable. However if ``fail_gracefully``
    is ``True``, function returns normally and it is up to the caller
    to implement some sort of fallback mechanism.

    :param service_proxy: nameko service proxy provided by ``RPCProxy``
    :param fail_gracefully: if ``True``, don't raise ``ServiceTimeout``
    :param timeout: timeout in seconds
    :param method_name: which method to call to check if service is healthy
    :raises: ServiceTimeout: if service is unresponsive and
        ``fail_gracefully`` is False
    :return: ``True`` if service is responsive
    """
    try:
        with eventlet.Timeout(timeout, ServiceTimeout):
            method = getattr(service_proxy, method_name)
            method()
            return True
    except ServiceTimeout:
        message = f"{service_proxy.service_name} is unreachable after {timeout} seconds"
        logger.warning(message)
        if fail_gracefully:
            return False
        else:
            raise ServiceTimeout(message)



import pytest
from nameko import config
from nameko.rpc import RpcProxy, rpc
from nameko.testing.services import entrypoint_hook

from nameko_chassis.health import is_service_responsive


@pytest.fixture(autouse=True)
def rabbit_config():
    with config.patch({"AMQP_URI": "pyamqp://localhost:5672"}):
        yield


class MyService:
    name = "my_service"
    my_service = RpcProxy(name)

    @rpc
    def up(self):
        return "ok"

    @rpc
    def is_responsive(self):
        return is_service_responsive(self.my_service, method_name="up")


def test_is_service_responsive(container_factory):
    container = container_factory(MyService)
    container.start()
    with entrypoint_hook(container, "is_responsive") as hook:
        assert hook() is True



import logging

import pytest
from nameko import config
from nameko.testing.services import entrypoint_hook
from nameko_prometheus.utils import reset_prometheus_registry
from werkzeug.test import EnvironBuilder

from nameko_chassis.service import Service


@pytest.fixture(autouse=True)
def rabbit_config():
    with config.patch({"AMQP_URI": "pyamqp://localhost:5672"}):
        yield


@pytest.fixture(autouse=True)
def reset_registry():
    reset_prometheus_registry("my_service")


class MyService(Service):
    name = "my_service"


def test_say_hello(container_factory):
    container = container_factory(MyService)
    container.start()
    with entrypoint_hook(container, "say_hello") as hook:
        assert hook() == "Hello from my_service!"


def test_serve_metrics(container_factory):
    request = EnvironBuilder(
        method="GET", path="/metrics", headers={"Accept": "text/html"}
    ).get_request()
    container = container_factory(MyService)
    container.start()
    with entrypoint_hook(container, "serve_metrics") as hook:
        response = hook(request)
        assert response.status_code == 200


def test_set_log_level(container_factory):
    container = container_factory(MyService)
    container.start()
    logger = logging.getLogger("foo.bar")
    logger.setLevel(logging.ERROR)
    with entrypoint_hook(container, "set_log_level") as hook:
        hook("foo.bar", logging.DEBUG)
    logger = logging.getLogger("foo.bar")
    assert logger.level == logging.DEBUG
